/**
 * @file code.c
 * @brief Implementation of tensor code
 */

#include "bch.h"
#include "code.h"
#include "parameters.h"
#include "repetition.h"
#include <stdint.h>
#include <string.h>
#ifdef VERBOSE
#include <stdio.h>
#include "vector.h"
#endif


/**
 *
 * @brief Encoding the message m to a code word em using the tensor code
 *
 * First we encode the message using the BCH code, then with the repetition code to obtain
 * a tensor code word.
 *
 * @param[out] em Pointer to an array that is the tensor code word
 * @param[in] m Pointer to an array that is the message
 */
void code_encode(uint64_t *em, const uint64_t *m) {

	uint64_t tmp[VEC_N1_SIZE_64] = {0};

	bch_code_encode(tmp, m);
	repetition_code_encode(em, tmp);

	#ifdef VERBOSE
		printf("\n\nBCH code word: "); vect_print(tmp, VEC_N1_SIZE_BYTES);
		printf("\n\nTensor code word: "); vect_print(em, VEC_N1N2_SIZE_BYTES);
	#endif
}



/**
 * @brief Decoding the code word em to a message m using the tensor code
 *
 * @param[out] m Pointer to an array that is the message
 * @param[in] em Pointer to an array that is the code word
 */
void code_decode(uint64_t *m, const uint64_t *em) {

	uint64_t tmp[VEC_N1_SIZE_64] = {0};

	repetition_code_decode(tmp, em);
	bch_code_decode(m, tmp);

	#ifdef VERBOSE
		printf("\n\nRepetition decoding result (the input for the BCH decoding algorithm): "); vect_print(tmp, VEC_N1_SIZE_BYTES);
	#endif

}
